package event_tracker

import (
	"fmt"
	"sync"
)

const (
	ModuleName = "event_tracker"
)

type Event struct {
	UserId    int64
	ProjectId int64
}

type Events struct {
	mu     sync.Mutex
	events []Event
}

type EventsInterface interface {
	EmitEvent(userId, projectId int64)
}

type EventTracker struct {
	eventLists map[string]*Events
}

func NewEventTracker() *EventTracker {
	return &EventTracker{
		eventLists: map[string]*Events{},
	}
}

type EventTrackerRegisterer interface {
	RegisterEvent(name string) EventsInterface
}

func (et *EventTracker) RegisterEvent(name string) EventsInterface {
	if _, exists := et.eventLists[name]; exists {
		panic(fmt.Errorf("event with name %s already exists", name))
	}
	e := &Events{}
	et.eventLists[name] = e
	return e
}

func (e *Events) EmitEvent(userId, projectId int64) {
	event := Event{
		UserId:    userId,
		ProjectId: projectId,
	}

	e.mu.Lock()
	defer e.mu.Unlock()

	e.events = append(e.events, event)
}

type EventTrackerCollector interface {
	CloneEventData() *EventData
	Subtract(*EventData)
}

func (et *EventTracker) CloneEventData() *EventData {
	events := make(map[string][]Event)
	for k, v := range et.eventLists {
		v.mu.Lock()
		if len(v.events) == 0 {
			v.mu.Unlock()
			continue
		}
		var newEvents []Event
		newEvents = append(newEvents, v.events...)
		events[k] = newEvents
		v.mu.Unlock()
	}
	return &EventData{
		Events: events,
	}
}

func (et *EventTracker) Subtract(ed *EventData) {
	for k, v := range ed.Events {
		e := et.eventLists[k]
		e.mu.Lock()
		e.events = append(e.events[:0], e.events[len(v):]...)
		e.mu.Unlock()
	}
}

// EventData is a struct that contains all the eventLists and used to send them to GitLab
type EventData struct {
	Events map[string][]Event
}

func (ed *EventData) IsEmpty() bool {
	return len(ed.Events) == 0
}

type EventTrackerInterface interface {
	EventTrackerRegisterer
	EventTrackerCollector
}
