package server

import (
	"context"
	"time"

	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/gitlab"
	gapi "gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/gitlab/api"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/module/event_tracker"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/module/modserver"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/module/modshared"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/tool"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/tool/errz"
	"go.uber.org/zap"
)

type module struct {
	log                  *zap.Logger
	api                  modserver.Api
	eventTracker         event_tracker.EventTrackerCollector
	gitLabClient         gitlab.ClientInterface
	eventReportingPeriod time.Duration
}

func (m *module) Run(ctx context.Context) error {
	if m.eventReportingPeriod == 0 {
		return nil
	}

	ticker := time.NewTicker(m.eventReportingPeriod)
	defer ticker.Stop()
	done := ctx.Done()
	for {
		select {
		case <-done:
			ctxExit, cancel := context.WithTimeout(context.Background(), 5*time.Second)
			defer cancel()
			// Flush events before exiting
			m.sendEvents(ctxExit) // nolint: contextcheck
			return nil
		case <-ticker.C:
			m.sendEvents(ctx)
		}
	}
}

func (m *module) sendEvents(ctx context.Context) {
	if err := m.sendEventsInternal(ctx); err != nil {
		if !errz.ContextDone(err) {
			m.api.HandleProcessingError(ctx, m.log, modshared.NoAgentId, "Failed to send event data", err)
		}
	}
}

func (m *module) sendEventsInternal(ctx context.Context) error {
	ed := m.eventTracker.CloneEventData()
	if ed.IsEmpty() {
		return nil
	}
	gapiEd := convertEvents(ed)
	if err := gapi.SendEvent(ctx, m.gitLabClient, gapiEd); err != nil {
		return err
	}
	m.eventTracker.Subtract(ed)
	return nil
}

func (m *module) Name() string {
	return event_tracker.ModuleName
}

func convertEvents(ed *event_tracker.EventData) gapi.EventData {
	em := tool.TransformMap[string, []event_tracker.Event, []gapi.Event, map[string][]event_tracker.Event, map[string][]gapi.Event](
		ed.Events,
		func(events []event_tracker.Event) int {
			return len(events)
		},
		func(events []event_tracker.Event, sink []gapi.Event) []gapi.Event {
			for _, e := range events {
				sink = append(sink, gapi.Event{
					UserId:    e.UserId,
					ProjectId: e.ProjectId,
				})
			}
			return sink
		},
		func(events []gapi.Event) []gapi.Event {
			return events
		},
	)
	return gapi.EventData{
		Events: em,
	}
}
